"""Maybe functor, applicative, monad, and alternative types."""

from typing import Callable, Generic, Optional, Sequence, TypeVar

import attr

A = TypeVar('A')
B = TypeVar('B')


@attr.s(frozen=True, slots=True)
class Maybe(Generic[A]):
    """Maybe sum type."""

    def __attrs_post_init__(self) -> None:
        # pylint: disable=no-self-use
        """Disallow direct construction."""
        raise TypeError()

    def fmap(self, f: Callable[[A], B]) -> 'Maybe[B]':
        """Stub the functor fmap operator."""
        raise NotImplementedError()

    def apply(self, f: 'Maybe[Callable[[A], B]]') -> 'Maybe[B]':
        """Stub the applicative apply operator."""
        raise NotImplementedError()

    def bind(self, f: Callable[[A], 'Maybe[B]']) -> 'Maybe[B]':
        """Stub the monad bind operator."""
        raise NotImplementedError()

    def then(self, other: 'Maybe[B]') -> 'Maybe[B]':
        # pylint: disable=no-self-use
        """Implement the monad then operator."""
        return other

    def fail(self, e: Exception) -> A:
        """Stub the monad fail operator."""
        raise NotImplementedError()

    def choice(self, a: 'Maybe[A]') -> 'Maybe[A]':
        """Stub the alternative choice operator."""
        raise NotImplementedError()

    def from_maybe(self, a: A) -> A:
        """Stub from_maybe."""
        raise NotImplementedError()

    def to_optional(self) -> Optional[A]:
        """Stub to_optional."""
        raise NotImplementedError()

    @classmethod
    def from_optional(cls, a: 'Optional[A]') -> 'Maybe[A]':
        """Return a maybe from an optional."""
        if a is None:
            return Nothing()
        return Just(a)

    def combine(
            self, f: Callable[[A], Callable[[A], A]], other: 'Maybe[A]'
    ) -> 'Maybe[A]':
        """Combine maybes."""
        return (
            other.apply(self.fmap(f))
            .choice(self)
            .choice(other)
        )

    @classmethod
    def cat(cls, xs: 'Sequence[Maybe[A]]') -> Sequence[A]:
        """Return a sequence of all the Just values."""
        return [x.value for x in xs if isinstance(x, Just)]


@attr.s(frozen=True, slots=True)
class Just(Generic[A], Maybe[A]):
    """Just product type."""

    value: A = attr.ib()

    def __attrs_post_init__(self) -> None:
        # pylint: disable=no-self-use
        """Allow direct construction."""

    def fmap(self, f: Callable[[A], B]) -> 'Maybe[B]':
        """Implement the functor fmap operator."""
        return Just(f(self.value))

    def apply(self, f: 'Maybe[Callable[[A], B]]') -> 'Maybe[B]':
        """Implement the applicative apply operator."""
        if isinstance(f, Nothing):
            return Nothing()
        if isinstance(f, Just):
            return self.fmap(f.value)
        raise TypeError()

    def bind(self, f: Callable[[A], 'Maybe[B]']) -> 'Maybe[B]':
        """Implement the monad bind operator."""
        return f(self.value)

    def fail(self, _: Exception) -> A:
        """Implement the monad fail operator."""
        return self.value

    def choice(self, a: 'Maybe[A]') -> 'Maybe[A]':
        """Implement the alternative choice operator."""
        return Just(self.value)

    def from_maybe(self, _: A) -> A:
        """Unwrap a maybe."""
        return self.value

    def to_optional(self) -> Optional[A]:
        """Return a maybe from an optional."""
        return self.value


@attr.s(frozen=True, slots=True)
class Nothing(Generic[A], Maybe[A]):
    """Nothing product type."""

    def __attrs_post_init__(self) -> None:
        """Allow direct construction."""

    def fmap(self, _: Callable[[A], B]) -> 'Nothing[B]':
        # pylint: disable=no-self-use
        """Implement the functor fmap operator."""
        return Nothing()

    def apply(self, _: 'Maybe[Callable[[A], B]]') -> 'Maybe[B]':
        # pylint: disable=no-self-use
        """Implement the applicative apply operator."""
        return Nothing()

    def bind(self, _: Callable[[A], 'Maybe[B]']) -> 'Maybe[B]':
        # pylint: disable=no-self-use
        """Implement the monad bind operator."""
        return Nothing()

    def fail(self, e: Exception) -> A:
        """Implement the monad fail operator."""
        raise e

    def choice(self, a: 'Maybe[A]') -> 'Maybe[A]':
        """Implement the alternative choice operator."""
        return attr.evolve(a)

    def from_maybe(self, a: A) -> A:
        """Unwrap a maybe."""
        return a

    def to_optional(self) -> Optional[A]:
        """Return a maybe from an optional."""
        return None



"""Build script for the C extension."""

import os
import re
from typing import Any, MutableMapping

from setuptools import Extension

NETSNMP_BASEDIR = os.path.join(
    os.path.dirname(os.path.realpath(__file__)), 'libs'
)
NETSNMP_LIBS = os.popen('net-snmp-config --libs').read()

EXTENSIONS = [
    Extension(
        'snmp_fetch.capi', [
            'src/capi/capimodule.cpp'
        ],
        depends=[
            'src/capi/session.hpp',
            'src/capi/asyncio.hpp',
            'src/capi/results.hpp',
            'src/capi/types.hpp',
            'src/capi/utils.hpp',
            'src/capi/debug.hpp',
        ],
        library_dirs=re.findall(r'-L(\S+)', NETSNMP_LIBS),
        libraries=re.findall(r'-l(\S+)', NETSNMP_LIBS),
        language='c++',
        extra_compile_args=['-std=c++17']
    ),
]


def build(setup_kwargs: MutableMapping[Any, Any]) -> None:
    """Entry point for build script."""
    setup_kwargs.update({'ext_modules': EXTENSIONS})



"""Functional programming structures."""

from functools import partial
from typing import Any, Callable, Sequence, TypeVar

from snmp_fetch.fp import either, maybe

__all__ = ['either', 'maybe']

A = TypeVar('A')
B = TypeVar('B')
C = TypeVar('C')

T = TypeVar('T')

F = partial


def curry2(f: Callable[[A, B], C]) -> Callable[[A], Callable[[B], C]]:
    """Curry a two argument function."""
    return lambda a: lambda b: f(a, b)


def star(
        f: Callable[..., T], *args: Any, **kwargs: Any
) -> Callable[[Sequence[Any]], T]:
    """Return a partial function that apply args with the star operator."""
    def apply(params: Sequence[Any]) -> T:
        return f(*params, *args, **kwargs)
    return apply



"""Utility functions."""

import re
from typing import Sequence, Text, Union, overload


def align64(x: int) -> int:
    """Align a size to be a multiple of 64."""
    if x < 0 or x >= 2 ** 64:
        raise ValueError(
            'integer does not fall between (' + str((2 ** 64) - 1) + ')'
        )
    return (x + 7) & ~0x07


@overload
def convert_oid(oid: Text) -> Sequence[int]:
    # pylint: disable=unused-argument
    # pragma: no cover
    """Convert a text oid to a sequence of integers."""
    ...  # pragma: no cover


@overload
def convert_oid(oid: Sequence[int]) -> Text:
    # pylint: disable=function-redefined, unused-argument
    # pragma: no cover
    """Convert a sequence of integers to a text oid."""
    ...  # pragma: no cover


def convert_oid(
        oid: Union[Text, Sequence[int]]
) -> Union[Sequence[int], Text]:
    # pylint: disable=function-redefined
    """Convert an oid between text and sequence of integers."""
    if isinstance(oid, Text):
        if re.match(r'^\.?\d+(\.\d+)*$', oid):
            if oid.startswith('.'):
                oid = oid[1:]
            return [int(x) for x in (oid).split('.')]
        raise ValueError(f'{oid} is not a valid oid')
    return '.'+'.'.join(map(str, oid))


@overload
def validate_oid(oid: Text) -> Text:
    # pylint: disable=unused-argument
    # pragma: no cover
    """Validate a text oid."""
    ...  # pragma: no cover


@overload
def validate_oid(oid: Sequence[int]) -> Sequence[int]:
    # pylint: disable=function-redefined, unused-argument
    """Validate a sequence of integers oid."""
    ...  # pragma: no cover


def validate_oid(
        oid: Union[Text, Sequence[int]]
) -> Union[Text, Sequence[int]]:
    """Validate an oid."""
    # pylint: disable=function-redefined
    return convert_oid(convert_oid(oid))



"""Stub file for C API."""

from typing import Optional, Sequence, Text, Tuple

import numpy as np


class ErrorType(type):
    """ErrorType stub."""

    SESSION_ERROR: 'ErrorType'
    CREATE_REQUEST_PDU_ERROR: 'ErrorType'
    SEND_ERROR: 'ErrorType'
    BAD_RESPONSE_PDU_ERROR: 'ErrorType'
    TIMEOUT_ERROR: 'ErrorType'
    ASYNC_PROBE_ERROR: 'ErrorType'
    TRANSPORT_DISCONNECT_ERROR: 'ErrorType'
    CREATE_RESPONSE_PDU_ERROR: 'ErrorType'


class SnmpError:
    # pylint: disable=too-few-public-methods
    """SnmpError stub."""

    type: ErrorType
    host: Tuple[int, Text, Text]
    sys_errno: Optional[int]
    snmp_errno: Optional[int]
    err_stat: Optional[int]
    err_index: Optional[int]
    err_oid: Optional[Sequence[int]]
    message: Optional[Text]

    retries: int
    timeout: int
    max_active_sessions: int
    max_var_binds_per_pdu: int
    max_bulk_repetitions: int

    def __init__(
            self,
            type: ErrorType,
            host: Tuple[int, Text, Text],
            sys_errno: Optional[int] = ...,
            snmp_errno: Optional[int] = ...,
            err_stat: Optional[int] = ...,
            err_index: Optional[int] = ...,
            err_oid: Optional[Sequence[int]] = ...,
            message: Optional[Text] = ...
    ) -> None:
        # pylint: disable=too-many-arguments, unused-argument, redefined-builtin
        """Initialize an SNMP config object."""
        ...


class PduType(type):
    """PduType stub."""

    GET_REQUEST: 'PduType'
    NEXT_REQUEST: 'PduType'
    BULKGET_REQUEST: 'PduType'


class SnmpConfig:
    # pylint: disable=too-few-public-methods
    """SnmpConfig stub."""

    retries: int
    timeout: int
    max_active_sessions: int
    max_var_binds_per_pdu: int
    max_bulk_repetitions: int

    def __init__(
            self,
            retries: int = ...,
            timeout: int = ...,
            max_active_sessions: int = ...,
            max_var_binds_per_pdu: int = ...,
            max_bulk_repetitions: int = ...
    ) -> None:
        # pylint: disable=too-many-arguments, unused-argument
        """Initialize an SNMP error object."""
        ...


def fetch(
        pdu_type: PduType,
        hosts: Sequence[Tuple[int, Text, Text]],
        var_binds: Sequence[Tuple[Sequence[int], Tuple[int, int]]],
        config: SnmpConfig = ...
) -> Tuple[Sequence[np.ndarray], Sequence[SnmpError]]:
    # pylint: disable=unused-argument
    """Fetch SNMP objects via the C API."""
    ...



"""Variable bindings."""

from operator import add
from typing import Any, Callable, Optional, Sequence, Text, Tuple, Union

import attr
import numpy as np
import pandas as pd
from toolz.functoolz import compose, identity

from snmp_fetch import dtype
from .fp import curry2
from .fp.maybe import Maybe
from .utils import convert_oid, validate_oid


def maybe_oid(oid: Optional[Text]) -> Maybe[Text]:
    """Validate and wrap an oid in a Maybe."""
    return Maybe.from_optional(oid).fmap(validate_oid)


@attr.s(frozen=True, slots=True)
class var_bind:
    # pylint: disable=invalid-name
    """Variable binding."""

    oid: Maybe[Text] = attr.ib(
        default=None, converter=maybe_oid
    )
    index: Maybe[np.dtype] = attr.ib(
        default=None, converter=dtype.to_dtype
    )
    data: Maybe[np.dtype] = attr.ib(
        default=None, converter=dtype.to_dtype
    )
    op: Callable[[Any], Any] = attr.ib(
        default=identity
    )

    def __lshift__(self, other: 'var_bind') -> 'var_bind':
        # pylint: disable=no-member
        """Combine variable bindings."""
        return var_bind(
            oid=(
                self.oid
                .combine(curry2(add), other.oid)
                .to_optional()
            ),
            index=(
                self.index
                .combine(
                    lambda x: lambda y: dtype.concatv(x, y).throw(),
                    other.index
                )
                .to_optional()
            ),
            data=(
                self.data
                .combine(
                    lambda x: lambda y: dtype.concatv(x, y).throw(),
                    other.data
                )
                .to_optional()
            ),
            op=compose(other.op, self.op)
        )

    header_cstruct = np.dtype([
        ('host_index', np.uint64),
        ('oid_size#', np.uint64),
        ('result_size#', np.uint64),
        ('result_type#', np.uint64),
        ('timestamp', 'datetime64[s]')
    ])

    def oid_cstruct(self) -> np.dtype:
        # pylint: disable=no-member
        """Get the oid cstruct."""
        return (
            self.oid
            .bind(lambda x: (
                dtype.array(np.dtype(np.uint64), len(convert_oid(x)))
                .fmap(lambda arr: np.dtype([('oid#', arr)]))
            ))
            .fail(AttributeError('oid has no dtype'))
        )

    def index_cstruct(self, pad: bool = True) -> Maybe[np.dtype]:
        # pylint: disable=no-member
        """Get the index cstruct with optional padding."""
        return(
            self.index
            .fmap(lambda x: dtype.pad64(x, 'index#').throw() if pad else x)
        )

    def data_cstruct(self, pad: bool = True) -> Maybe[np.dtype]:
        # pylint: disable=no-member
        """Get the data cstruct with optional padding."""
        return (
            self.data
            .fmap(lambda x: dtype.pad64(x, 'data#').throw() if pad else x)
        )

    def cstruct(self) -> np.dtype:
        # pylint: disable=no-member
        """Create the var_bind cstruct."""
        return dtype.concat([
            self.header_cstruct,
            self.oid_cstruct(),
            *Maybe.cat([
                self.index_cstruct(),
                self.data_cstruct()
            ])
        ]).throw()

    def null_cstruct(
            self, param: Optional[Text] = None
    ) -> Tuple[Sequence[int], Tuple[int, int]]:
        # pylint: disable=no-member
        """Return a null variable binding cstruct with optional parameter."""
        return convert_oid(
            self.oid
            .combine(curry2(add), maybe_oid(param))
            .fail(AttributeError('var_bind has no oid'))
        ), (
            self.oid_cstruct().itemsize + (
                self.index_cstruct(pad=False)
                .fmap(lambda x: x.itemsize)
                .from_maybe(0)
            ), (
                self.data_cstruct(pad=False)
                .fmap(lambda x: x.itemsize)
                .from_maybe(0)
            )
        )

    def __call__(
            self, param: Optional[Text] = None
    ) -> Tuple[Sequence[int], Tuple[int, int]]:
        """Return a null variable binding cstruct with optional parameter."""
        return self.null_cstruct(param)

    def view(self, arr: np.ndarray) -> Any:
        # pylint: disable=no-member
        """Convert a cstruct array to a dataframe."""
        view = arr.view(self.cstruct())
        df = pd.DataFrame.from_records(
            view.tolist(), columns=view.dtype.names
        )
        df = df[~df['result_type#'].isin([128, 129, 130])]
        df = self.op(df)  # pylint: disable=not-callable
        df = df.drop(columns=({
            'oid#', 'oid_size#', 'result_size#', 'result_type#', 'index#',
            'data#', 'mask'
        }.intersection(df.columns)))
        df['timestamp'] = df['timestamp'].dt.tz_localize('UTC')
        return df


def merge_results(
        results: Sequence[np.ndarray], var_binds: Sequence[var_bind],
        extras: Optional[Any] = None,
        index: Union[Text, Sequence[Text]] = 'host_index'
) -> Optional[Any]:
    """Merge result variable bindings."""
    df = None
    for v, result in zip(var_binds, results):
        if result.size == 0:
            continue
        _df = v.view(result)
        if df is None:
            df = _df
        else:
            df = df.merge(
                _df, how='outer', left_index=True, right_index=True,
                sort=True
            )
            # Hitting an issue when a row enters/leaves mid collection which
            # results in an NaT in the timestamp column for the column missing
            # row data on merge.  Running max with the NaT value causes the
            # entire timestamp column to fill with NaN.  To avoid this, NaT's
            # are filled with the max timestamp value in the respective column
            # prior to getting the max value.
            for tkey in ['timestamp_x', 'timestamp_y']:
                df[tkey] = df[tkey].fillna(df[tkey].max())
            df['timestamp'] = df[['timestamp_x', 'timestamp_y']].max(axis=1)
            df = df.drop(columns=['timestamp_x', 'timestamp_y'])
    if df is not None:
        df = df.reset_index()
        if extras is not None:
            df = (
                df.set_index(index)
                .merge(
                    extras.set_index(index), how='left', left_index=True,
                    right_index=True
                )
                .reset_index()
                .drop(columns='host_index')
            )
    return df



"""Maybe functor, applicative, and monad types."""

from typing import Callable, Generic, TypeVar

import attr

A = TypeVar('A')
B = TypeVar('B')
C = TypeVar('C')


@attr.s(frozen=True, slots=True)
class Either(Generic[A, B]):
    """Either sum type."""

    def __attrs_post_init__(self) -> None:
        # pylint: disable=no-self-use
        """Disallow direct construction."""
        raise TypeError()

    def fmap(self, f: Callable[[B], C]) -> 'Either[A, C]':
        """Stub the functor fmap operator."""
        raise NotImplementedError()

    def apply(self, f: 'Either[A, Callable[[B], C]]') -> 'Either[A, C]':
        """Implement the applicative apply operator."""
        if isinstance(f, Left):
            return Left(f.value)
        if isinstance(f, Right):
            return self.fmap(f.value)
        raise TypeError()

    def bind(self, f: 'Callable[[B], Either[A, C]]') -> 'Either[A, C]':
        """Stub the monad bind operator."""
        raise NotImplementedError()

    def then(self, other: 'Either[A, C]') -> 'Either[A, C]':
        # pylint: disable=no-self-use
        """Implement the monad then operator."""
        return other

    def from_left(self, _: A) -> A:
        """Stub from_left."""
        raise NotImplementedError()

    def from_right(self, _: B) -> B:
        """Stub from_right."""
        raise NotImplementedError()

    def throw(self) -> B:
        """Stub throw."""
        raise NotImplementedError()


@attr.s(frozen=True, slots=True)
class Left(Generic[A, B], Either[A, B]):
    """Left product type."""

    value: A = attr.ib()

    def __attrs_post_init__(self) -> None:
        # pylint: disable=no-self-use
        """Allow direct construction."""

    def fmap(self, f: Callable[[B], C]) -> 'Either[A, C]':
        """Implement the functor fmap operator."""
        return Left(self.value)

    def bind(self, f: 'Callable[[B], Either[A, C]]') -> 'Either[A, C]':
        """Implement the monad bind operator."""
        return Left(self.value)

    def from_left(self, _: A) -> A:
        """Unwrap a left."""
        return self.value

    def from_right(self, b: B) -> B:
        """Return the default value."""
        return b

    def throw(self) -> B:
        # pylint: disable=raising-non-exception
        """Implement throw."""
        if isinstance(self.value, Exception):
            raise self.value
        raise TypeError(f'{self.value} is not an exception.')


@attr.s(frozen=True, slots=True)
class Right(Generic[A, B], Either[A, B]):
    """Right product type."""

    value: B = attr.ib()

    def __attrs_post_init__(self) -> None:
        """Allow construction."""

    def fmap(self, f: Callable[[B], C]) -> 'Either[A, C]':
        # pylint: disable=no-self-use
        """Implement the functor fmap operator."""
        return Right(f(self.value))

    def bind(self, f: 'Callable[[B], Either[A, C]]') -> 'Either[A, C]':
        """Implement the monad bind operator."""
        return f(self.value)

    def from_left(self, a: A) -> A:
        """Return the default value."""
        return a

    def from_right(self, _: B) -> B:
        """Unwrap a right."""
        return self.value

    def throw(self) -> B:
        """Unwrap a right."""
        return self.value



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['snmp_fetch', 'snmp_fetch.fp']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=19.1,<20.0',
 'numpy>=1.16,<2.0',
 'pandas>=0.25,<0.26',
 'toolz>=0.10.0,<0.11.0']

setup_kwargs = {
    'name': 'snmp-fetch',
    'version': '0.1.4',
    'description': 'An opinionated python SNMPv2 library built for rapid database ingestion.',
    'long_description': 'snmp-fetch\n==========\n\n|Version badge| |Python version badge| |PyPI format badge| |Build badge| |Coverage badge|\n\n.. |Version badge| image:: https://img.shields.io/pypi/v/snmp-fetch\n   :target: https://pypi.org/project/snmp-fetch/\n\n.. |Python version badge| image:: https://img.shields.io/pypi/pyversions/snmp-fetch\n   :alt: PyPI - Python Version\n   :target: https://pypi.org/project/snmp-fetch/\n  \n.. |PyPI format badge| image:: https://img.shields.io/pypi/format/snmp-fetch\n   :alt: PyPI - Format\n   :target: https://pypi.org/project/snmp-fetch/\n\n.. |Build badge| image:: https://travis-ci.org/higherorderfunctor/snmp-fetch.svg?branch=master\n   :target: https://travis-ci.org/higherorderfunctor/snmp-fetch\n\n.. |Coverage badge| image:: https://coveralls.io/repos/github/higherorderfunctor/snmp-fetch/badge.svg\n   :target: https://coveralls.io/github/higherorderfunctor/snmp-fetch\n\nAn opinionated python3.7 SNMPv2 library designed for rapid database ingestion.\n\nPrerequisites\n"""""""""""""\n\nSnmp-fetch is built for python 3.7 and c++17.  Building is currently only tested on gcc 8.   The following prerequisites must also be installed before adding snmp-fetch to your project.\n\nnet-snmp\n\'\'\'\'\'\'\'\'\n\nA recent version of net-snmp is required; testing has only been performed against net-snmp 5.8.\n\n.. code:: console\n\n   # perl bindings may be needed on RPM based systems\n   sudo yum install perl-devel\n\n   # compile and install net-snmp\n   wget https://sourceforge.net/projects/net-snmp/files/net-snmp/5.8/net-snmp-5.8.tar.gz/download -O net-snmp.tar.gz\n   tar xzfv net-snmp.tar.gz\n   cd net-snmp-5.8\n   ./configure --enable-ipv6 --with-defaults\n   make\n   sudo make install\n   cd .. && rm -rf net-snmp*\n   sudo ldconfig -v\n\nboost (headers only)\n\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\n\nBoost is a popular C++ library to reduce boilerplate.  Snmp-fetch makes use of some of the header only libraries.\n\n.. code:: console\n\n   wget https://dl.bintray.com/boostorg/release/1.70.0/source/boost_1_70_0.tar.gz\n   tar xzfv boost_1_70_0.tar.gz\n   sudo mv boost_1_70_0/boost /usr/local/include/\n   rm -rf boost_1_70_0*\n\npybind11\n\'\'\'\'\'\'\'\'\n\nPybind11 is a C++ wrapper around the Python C API to reduce boilerplate.  This is a header only library, but the test script will attempt to build a binary.\n\n.. code:: console\n\n   # install pytest in the user space for the test build\n   pip3.7 install --user pytest\n\n   # install cmake\n   wget https://github.com/Kitware/CMake/releases/download/v3.15.1/cmake-3.15.1-Linux-x86_64.sh\n   sudo ./cmake-3.15.1-Linux-x86_64.sh \\\n     --prefix=/usr/local/ \\\n     --exclude-subdir \\\n     --skip-license\n   rm cmake-3.15.1-Linux-x86_64.sh\n\n   # test and install pybind11\n   git clone --depth 1 https://github.com/pybind/pybind11.git\n   cd pybind11\n   mkdir -p build && cd build\n   cmake .. -DPYBIND11_CPP_STANDARD=-std=c++17 -DDOWNLOAD_CATCH=1\n   make check -j 4\n   sudo make install\n   cd ../../\n   rm -rf pybind11\n\nInstallation\n""""""""""""\n\n.. code:: console\n\n   # poetry\n   poetry add snmp-fetch --git https://gitlab.com/higherorderfunctor/snmp-fetch.git\n   # pip\n   pip install git+https://gitlab.com/higherorderfunctor/snmp-fetch.git\n\nDevelopment\n"""""""""""\n\n.. code:: console\n\n   # poetry must be installed\n   git clone ...\n   cd snmp-fetch\n   virtualenv -p python3.7 ENV\n   source ENV/bin/activate\n   poetry install  # if never installed\n   deactivate && source ENV/bin/activate  # refresh PATH\n\n   # fast fail testing\n   pytest --hypothesis-show-statistics -x --ff\n\n   # testing\n   coverage erase\n   pytest --cov --hypothesis-show-statistics\n   coverage html\n\n   # linting\n   pylint *\n   flake8\n   mypy -p snmp_fetch -p tests\n   bandit -r snmp_fetch\n\n   # clean up imports\n   isort -rc --atomic .\n\nKnown Limitations\n"""""""""""""""""\n\n- The library only supports SNMPv2 at this time.\n\n- `BULKGET_REQUEST` and `NEXT_REQUEST` will always perform a walk.\n\n- Walks will always end if the root of the oid runs past the requested oid.\n\n- Duplicate objects on the same host/request will be silently discarded.\n\n  - This includes the initial request; walks must be performed on an oid prior to the first desired.\n\n  - ENDOFMIBVIEW variable bindings are silently discarded for this same reason.\n',
    'author': 'Christopher Aubut',
    'author_email': 'christopher@aubut.me',
    'url': 'https://github.com/higherorderfunctor/snmp-fetch',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}
from build import *
build(setup_kwargs)

setup(**setup_kwargs)



"""Datatype wrappers and helper functions."""

import ipaddress as ip
from operator import add
from typing import Any, Callable, Optional, Sequence, Text, Tuple, TypeVar, Union

import numpy as np

from .fp import F as Fn
from .fp.either import Either, Left, Right
from .fp.maybe import Just, Maybe, Nothing
from .utils import align64

FuncType = Callable[..., Any]
F = TypeVar('F', bound=FuncType)
T = TypeVar('T')

DtypeStruct = Sequence[Tuple[Text, np.dtype]]


# monkey patch classes from the ipaddress module to support comparing IPv4 and
# IPv6 objects.

def monkeypatch(cls: type, method: Text) -> Callable[[F], F]:
    """Monkey patch and store base method on the function object."""
    def patch(f: F) -> F:
        if not getattr(cls, method) == f:
            setattr(f, method, getattr(cls, method))
            setattr(cls, method, f)
        return f
    return patch


@monkeypatch(ip.IPv4Address, '__lt__')
def ipv4_address__lt__(self: T, other: T) -> bool:
    # pylint: disable=no-member
    """Compare less than on mixed IP address object types."""
    if isinstance(other, ip.IPv4Address):
        return ipv4_address__lt__.__lt__(self, other)  # type: ignore
    if isinstance(other, ip.IPv6Address):
        return True
    raise TypeError(f'{other} is not an IPv4Address or IPv6Address object')


@monkeypatch(ip.IPv4Address, '__le__')
def ipv4_address__le__(self: T, other: T) -> bool:
    """Compare less than or equal to on mixed IP address object types."""
    if isinstance(other, ip.IPv4Address):
        return ipv4_address__le__.__le__(self, other)  # type: ignore
    if isinstance(other, ip.IPv6Address):
        return True
    raise TypeError(f'{other} is not an IPv4Address or IPv6Address object')


@monkeypatch(ip.IPv4Address, '__gt__')
def ipv4_address__gt__(self: T, other: T) -> bool:
    """Compare greater than on mixed IP address object types."""
    if isinstance(other, ip.IPv4Address):
        return ipv4_address__gt__.__gt__(self, other)  # type: ignore
    if isinstance(other, ip.IPv6Address):
        return False
    raise TypeError(f'{other} is not an IPv4Address or IPv6Address object')


@monkeypatch(ip.IPv4Address, '__ge__')
def ipv4_address__ge__(self: T, other: T) -> bool:
    """Compare greater than or equal to on mixed IP address object types."""
    if isinstance(other, ip.IPv4Address):
        return ipv4_address__ge__.__ge__(self, other)  # type: ignore
    if isinstance(other, ip.IPv6Address):
        return False
    raise TypeError(f'{other} is not an IPv4Address or IPv6Address object')


@monkeypatch(ip.IPv6Address, '__lt__')
def ipv6_address__lt__(self: T, other: T) -> bool:
    """Compare less than on mixed IP address object types."""
    if isinstance(other, ip.IPv4Address):
        return False
    if isinstance(other, ip.IPv6Address):
        return ipv6_address__lt__.__lt__(self, other)  # type: ignore
    raise TypeError(f'{other} is not an IPv4Address or IPv6Address object')


@monkeypatch(ip.IPv6Address, '__le__')
def ipv6_address__le__(self: T, other: T) -> bool:
    """Compare less than or equal to on mixed IP address object types."""
    if isinstance(other, ip.IPv4Address):
        return False
    if isinstance(other, ip.IPv6Address):
        return ipv6_address__le__.__le__(self, other)  # type: ignore
    raise TypeError(f'{other} is not an IPv4Address or IPv6Address object')


@monkeypatch(ip.IPv6Address, '__gt__')
def ipv6_address__gt__(self: T, other: T) -> bool:
    """Compare greater than on mixed IP address object types."""
    if isinstance(other, ip.IPv4Address):
        return True
    if isinstance(other, ip.IPv6Address):
        return ipv6_address__gt__.__gt__(self, other)  # type: ignore
    raise TypeError(f'{other} is not an IPv4Address or IPv6Address object')


@monkeypatch(ip.IPv6Address, '__ge__')
def ipv6_address__ge__(self: T, other: T) -> bool:
    """Compare greater than or equal to on mixed IP address object types."""
    if isinstance(other, ip.IPv4Address):
        return True
    if isinstance(other, ip.IPv6Address):
        return ipv6_address__ge__.__ge__(self, other)  # type: ignore
    raise TypeError(f'{other} is not an IPv4Address or IPv6Address object')


@monkeypatch(ip.IPv4Network, '__lt__')
def ipv4_network__lt__(self: T, other: T) -> bool:
    """Compare less than on mixed IP network object types."""
    if isinstance(other, ip.IPv4Network):
        return ipv4_network__lt__.__lt__(self, other)  # type: ignore
    if isinstance(other, ip.IPv6Network):
        return True
    raise TypeError(f'{other} is not an IPv4Network or IPv6Network object')


@monkeypatch(ip.IPv4Network, '__le__')
def ipv4_network__le__(self: T, other: T) -> bool:
    """Compare less than or equal to on mixed IP network object types."""
    if isinstance(other, ip.IPv4Network):
        return ipv4_network__le__.__le__(self, other)  # type: ignore
    if isinstance(other, ip.IPv6Network):
        return True
    raise TypeError(f'{other} is not an IPv4Network or IPv6network object')


@monkeypatch(ip.IPv4Network, '__gt__')
def ipv4_network__gt__(self: T, other: T) -> bool:
    """Compare greater than on mixed IP network object types."""
    if isinstance(other, ip.IPv4Network):
        return ipv4_network__gt__.__gt__(self, other)  # type: ignore
    if isinstance(other, ip.IPv6Network):
        return False
    raise TypeError(f'{other} is not an IPv4Network or IPv6network object')


@monkeypatch(ip.IPv4Network, '__ge__')
def ipv4_network__ge__(self: T, other: T) -> bool:
    """Compare greater than or equal to on mixed IP network object types."""
    if isinstance(other, ip.IPv4Network):
        return ipv4_network__ge__.__ge__(self, other)  # type: ignore
    if isinstance(other, ip.IPv6Network):
        return False
    raise TypeError(f'{other} is not an IPv4Network or IPv6network object')


@monkeypatch(ip.IPv6Network, '__lt__')
def ip6_network__lt__(self: T, other: T) -> bool:
    """Compare less than on mixed IP network object types."""
    if isinstance(other, ip.IPv4Network):
        return False
    if isinstance(other, ip.IPv6Network):
        return ip6_network__lt__.__lt__(self, other)  # type: ignore
    raise TypeError(f'{other} is not an IPv4Network or IPv6network object')


@monkeypatch(ip.IPv6Network, '__le__')
def ipv6_network__le__(self: T, other: T) -> bool:
    """Compare less than or equal to on mixed IP network object types."""
    if isinstance(other, ip.IPv4Network):
        return False
    if isinstance(other, ip.IPv6Network):
        return ipv6_network__le__.__le__(self, other)  # type: ignore
    raise TypeError(f'{other} is not an IPv4Network or IPv6network object')


@monkeypatch(ip.IPv6Network, '__gt__')
def ipv6_network__gt__(self: T, other: T) -> bool:
    """Compare greater than on mixed IP network object types."""
    if isinstance(other, ip.IPv4Network):
        return True
    if isinstance(other, ip.IPv6Network):
        return ipv6_network__gt__.__gt__(self, other)  # type: ignore
    raise TypeError(f'{other} is not an IPv4Network or IPv6network object')


@monkeypatch(ip.IPv6Network, '__ge__')
def ipv6_network__ge__(self: T, other: T) -> bool:
    """Compare greater than or equal to on mixed IP network object types."""
    if isinstance(other, ip.IPv4Network):
        return True
    if isinstance(other, ip.IPv6Network):
        return ipv6_network__ge__.__ge__(self, other)  # type: ignore
    raise TypeError(f'{other} is not an IPv4Network or IPv6network object')


@monkeypatch(ip.IPv4Interface, '__lt__')
def ipv4_interface__lt__(self: T, other: T) -> bool:
    """Compare less than on mixed IP interface object types."""
    if isinstance(other, ip.IPv4Interface):
        return ipv4_interface__lt__.__lt__(self, other)  # type: ignore
    if isinstance(other, ip.IPv6Interface):
        return True
    raise TypeError(
        f'{other} is not an IPv4Interface or IPv6interface object'
    )


@monkeypatch(ip.IPv4Interface, '__le__')
def ipv4_interface__le__(self: T, other: T) -> bool:
    """Compare less than or equal to on mixed IP interface object types."""
    if isinstance(other, ip.IPv4Interface):
        return ipv4_interface__le__.__le__(self, other)  # type: ignore
    if isinstance(other, ip.IPv6Interface):
        return True
    raise TypeError(
        f'{other} is not an IPv4Interface or IPv6interface object'
    )


@monkeypatch(ip.IPv4Interface, '__gt__')
def ipv4_interface__gt__(self: T, other: T) -> bool:
    """Compare greater than on mixed IP interface object types."""
    if isinstance(other, ip.IPv4Interface):
        return ipv4_interface__gt__.__gt__(self, other)  # type: ignore
    if isinstance(other, ip.IPv6Interface):
        return False
    raise TypeError(
        f'{other} is not an IPv4Interface or IPv6interface object'
    )


@monkeypatch(ip.IPv4Interface, '__ge__')
def ipv4_interface__ge__(self: T, other: T) -> bool:
    """Compare greater than or equal to on mixed IP interface object types."""
    if isinstance(other, ip.IPv4Interface):
        return ipv4_interface__ge__.__ge__(self, other)  # type: ignore
    if isinstance(other, ip.IPv6Interface):
        return False
    raise TypeError(
        f'{other} is not an IPv4Interface or IPv6interface object'
    )


@monkeypatch(ip.IPv6Interface, '__lt__')
def ipv6_interface__lt__(self: T, other: T) -> bool:
    """Compare less than on mixed IP interface object types."""
    if isinstance(other, ip.IPv4Interface):
        return False
    if isinstance(other, ip.IPv6Interface):
        return ipv6_interface__lt__.__lt__(self, other)  # type: ignore
    raise TypeError(
        f'{other} is not an IPv4Interface or IPv6interface object'
    )


@monkeypatch(ip.IPv6Interface, '__le__')
def ipv6_interface__le__(self: T, other: T) -> bool:
    """Compare less than or equal to on mixed IP interface object types."""
    if isinstance(other, ip.IPv4Interface):
        return False
    if isinstance(other, ip.IPv6Interface):
        return ipv6_interface__le__.__le__(self, other)  # type: ignore
    raise TypeError(
        f'{other} is not an IPv4Interface or IPv6interface object'
    )


@monkeypatch(ip.IPv6Interface, '__gt__')
def ipv6_interface__gt__(self: T, other: T) -> bool:
    """Compare greater than on mixed IP interface object types."""
    if isinstance(other, ip.IPv4Interface):
        return True
    if isinstance(other, ip.IPv6Interface):
        return ipv6_interface__gt__.__gt__(self, other)  # type: ignore
    raise TypeError(
        f'{other} is not an IPv4Interface or IPv6interface object'
    )


@monkeypatch(ip.IPv6Interface, '__ge__')
def ipv6_interface__ge__(self: T, other: T) -> bool:
    """Compare greater than or equal to on mixed IP interface object types."""
    if isinstance(other, ip.IPv4Interface):
        return True
    if isinstance(other, ip.IPv6Interface):
        return ipv6_interface__ge__.__ge__(self, other)  # type: ignore
    raise TypeError(
        f'{other} is not an IPv4Interface or IPv6interface object'
    )


IpAddress = Union[ip.IPv4Address, ip.IPv6Address]
IpNetwork = Union[ip.IPv4Network, ip.IPv6Network]
IpInterface = Union[ip.IPv4Interface, ip.IPv6Interface]


def struct(ds: DtypeStruct) -> np.dtype:
    """Restructure a datatype."""
    return np.dtype(list(ds))


def destruct(ds: np.dtype) -> Either[Exception, DtypeStruct]:
    """Destructure a structured datatype."""
    if hasattr(ds, 'fields') and ds.fields:
        return Right([(f, d) for f, (d, *_) in ds.fields.items()])
    return Left(AttributeError(f'can only pad structured dtypes: {type(ds)}'))


def to_dtype(d: Optional[Union[np.dtype, DtypeStruct]]) -> Maybe[np.dtype]:
    """Convert varying forms of datatypes to a Maybe wrapped np.dtype."""
    if d is None:
        return Nothing()
    if isinstance(d, np.dtype):
        return Just(d)
    return Just(struct(d))


def concat(ds: Sequence[np.dtype]) -> Either[Exception, np.dtype]:
    """Concat structured datatypes."""
    def _concat(_ds: Sequence[np.dtype]) -> Either[Exception, DtypeStruct]:
        head, *tail = _ds
        if tail:
            return (
                destruct(head)
                .bind(lambda h: _concat(tail).fmap(Fn(add, h)))
            )
        return destruct(head)
    return (
        _concat(ds)
        .fmap(struct)
    )


def concatv(*ds: np.dtype) -> Either[Exception, np.dtype]:
    """Concat structured datatypes from a variable number of args."""
    return concat(ds)


def array(d: np.dtype, size: int) -> Maybe[np.dtype]:
    """Return an (n,) datatype if possible.

    Returns a bare datatype for a scalar request of Nothing with a negative
    size.
    """
    if size <= 0:
        return Nothing()
    if size == 1:
        return Just(d)
    return Just(np.dtype((d, size)))


def extend(ds: np.dtype, amount: int) -> np.dtype:
    """Extend the tail element of a structured datatype if possible.

    Currently only support character strings.
    """
    d = (
        destruct(ds)
        .fmap(lambda d: list(d[:-1]) + [(d[-1][0], extend(d[-1][1], amount))])
        .fmap(struct)
        .from_right(ds)
    )
    if d.kind == 'S':
        return np.dtype(f'S{d.itemsize + amount}')
    return d


def pad64(
        ds: np.dtype, label: Text = 'padding#'
) -> Either[Exception, np.dtype]:
    """Pad structured datatypes."""
    amount = align64(ds.itemsize) - ds.itemsize

    if amount == 0:
        return Right(ds)

    ds = extend(ds, amount)
    amount = align64(ds.itemsize) - ds.itemsize

    return (
        array(np.dtype(np.uint8), amount)
        .fmap(lambda d: concatv(ds, np.dtype([(label, d)])))
        .from_maybe(Right(ds))
    )


IPV4_PREFIX_LOOKUP_TABLE = {
    (0xFFFFFFFF << i) & 0xFFFFFFFF: 32 - i for i in range(0, 33)
}

IPV4_MASK_LOOKUP_TABLE = {
    v: k for k, v in IPV4_PREFIX_LOOKUP_TABLE.items()
}

IPV6_PREFIX_LOOKUP_TABLE = {
    (
        (0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF << i) &
        0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
    ): 128 - i
    for i in range(0, 129)
}

IPV6_MASK_LOOKUP_TABLE = {
    v: k for k, v in IPV6_PREFIX_LOOKUP_TABLE.items()
}


def ip_address(
        addr: bytes, zone: bool = False
) -> Union[IpAddress, Tuple[IpAddress, int]]:
    # pylint: disable=too-many-return-statements
    """Convert bytes to IP address."""
    if len(addr) == 4:
        if zone:
            return ip.IPv4Address(
                int.from_bytes(addr[:4], byteorder='big')
            ), -1
        return ip.IPv4Address(int.from_bytes(addr[:4], byteorder='big'))
    if len(addr) == 8:
        if zone:
            return (
                ip.IPv4Address(int.from_bytes(addr[:4], byteorder='big')),
                int.from_bytes(addr[-4:], byteorder='big')
            )
        return ip.IPv4Address(int.from_bytes(addr[:4], byteorder='big'))
    if len(addr) == 16:
        if zone:
            return ip.IPv6Address(
                int.from_bytes(addr[:16], byteorder='big')
            ), -1
        return ip.IPv6Address(int.from_bytes(addr[:16], byteorder='big'))
    if len(addr) == 20:
        if zone:
            return (
                ip.IPv6Address(int.from_bytes(addr[:16], byteorder='big')),
                int.from_bytes(addr[-4:], byteorder='big')
            )
        return ip.IPv6Address(int.from_bytes(addr[:16], byteorder='big'))
    raise TypeError(f'No matching address type for {addr}')


def ip_network(
        addr: bytes, mask: Union[bytes, int], zone: bool = False
) -> Union[IpNetwork, Tuple[IpNetwork, int]]:
    # pylint: disable=too-many-return-statements
    """Convert bytes to IP network."""
    if len(addr) == 4:
        if isinstance(mask, bytes):
            mask = IPV4_PREFIX_LOOKUP_TABLE[
                int.from_bytes(mask, byteorder='big')
            ]
        if zone:
            return ip.IPv4Network(
                (int.from_bytes(addr[:4], byteorder='big'), mask),
                strict=False
            ), -1
        return ip.IPv4Network(
            (int.from_bytes(addr[:4], byteorder='big'), mask), strict=False
        )
    if len(addr) == 8:
        if isinstance(mask, bytes):
            mask = IPV4_PREFIX_LOOKUP_TABLE[
                int.from_bytes(mask, byteorder='big')
            ]
        if zone:
            return ip.IPv4Network(
                (int.from_bytes(addr[:4], byteorder='big'), mask),
                strict=False
            ), int.from_bytes(addr[-4:], byteorder='big')
        return ip.IPv4Network(
            (int.from_bytes(addr[:4], byteorder='big'), mask), strict=False
        )
    if len(addr) == 16:
        if isinstance(mask, bytes):
            mask = IPV6_PREFIX_LOOKUP_TABLE[
                int.from_bytes(mask, byteorder='big')
            ]
        if zone:
            return ip.IPv6Network(
                (int.from_bytes(addr[:16], byteorder='big'), mask),
                strict=False
            ), -1
        return ip.IPv6Network(
            (int.from_bytes(addr[:16], byteorder='big'), mask), strict=False
        )
    if len(addr) == 20:
        if isinstance(mask, bytes):
            mask = IPV6_PREFIX_LOOKUP_TABLE[
                int.from_bytes(mask, byteorder='big')
            ]
        if zone:
            return ip.IPv6Network(
                (int.from_bytes(addr[:16], byteorder='big'), mask),
                strict=False
            ), int.from_bytes(addr[-4:], byteorder='big')
        return ip.IPv6Network(
            (int.from_bytes(addr[:16], byteorder='big'), mask), strict=False
        )
    raise TypeError(f'No matching network type for {addr}')


def ip_interface(
        addr: bytes, mask: Union[bytes, int], zone: bool = False
) -> Union[IpInterface, Tuple[IpInterface, int]]:
    # pylint: disable=too-many-return-statements
    """Convert bytes to IP interface."""
    if len(addr) == 4:
        if isinstance(mask, bytes):
            mask = IPV4_PREFIX_LOOKUP_TABLE[
                int.from_bytes(mask, byteorder='big')
            ]
        if zone:
            return ip.IPv4Interface(
                (int.from_bytes(addr[:4], byteorder='big'), mask)
            ), -1
        return ip.IPv4Interface(
            (int.from_bytes(addr[:4], byteorder='big'), mask)
        )
    if len(addr) == 8:
        if isinstance(mask, bytes):
            mask = IPV4_PREFIX_LOOKUP_TABLE[
                int.from_bytes(mask, byteorder='big')
            ]
        if zone:
            return ip.IPv4Interface(
                (int.from_bytes(addr[:4], byteorder='big'), mask)
            ), int.from_bytes(addr[-4:], byteorder='big')
        return ip.IPv4Interface(
            (int.from_bytes(addr[:4], byteorder='big'), mask)
        )
    if len(addr) == 16:
        if isinstance(mask, bytes):
            mask = IPV6_PREFIX_LOOKUP_TABLE[
                int.from_bytes(mask, byteorder='big')
            ]
        if zone:
            return ip.IPv6Interface(
                (int.from_bytes(addr[:16], byteorder='big'), mask)
            ), -1
        return ip.IPv6Interface(
            (int.from_bytes(addr[:16], byteorder='big'), mask)
        )
    if len(addr) == 20:
        if isinstance(mask, bytes):
            mask = IPV6_PREFIX_LOOKUP_TABLE[
                int.from_bytes(mask, byteorder='big')
            ]
        if zone:
            return ip.IPv6Interface(
                (int.from_bytes(addr[:16], byteorder='big'), mask)
            ), int.from_bytes(addr[-4:], byteorder='big')
        return ip.IPv6Interface(
            (int.from_bytes(addr[:16], byteorder='big'), mask)
        )
    raise TypeError(f'No matching interface type for {addr}')


def cast_ip_array(
        addr: np.ndarray, length: Optional[int] = None,
        ip_type: Optional[int] = None
) -> bytes:
    """Check IP types."""
    if length is None:
        length = len(addr)
    if ip_type is not None:
        if (ip_type, length) not in ((1, 4), (2, 16), (3, 8), (4, 20)):
            raise TypeError(
                f'No matching ip type for {(ip_type, length, addr)}'
            )
    return addr[:length].astype(np.uint8).tobytes()  # type: ignore
